package Undersampling


import org.apache.log4j.Logger
import org.apache.log4j.Level

import org.apache.spark.SparkContext
import org.apache.spark.SparkContext._
import org.apache.spark.SparkConf
import utils.KeelParser
import org.apache.spark.broadcast.Broadcast
import scala.collection.mutable.ListBuffer
import org.apache.mahout.keel.Dataset.InstanceSet
import org.apache.mahout.keel.Algorithms.ImbalancedClassification.Ensembles._
import java.io.IOException
import utils.Utils
import java.io.File
import org.apache.spark.mllib.regression.LabeledPoint
import org.apache.spark.mllib.linalg.Vectors
import java.util.ArrayList
import com.typesafe.config.ConfigFactory
import utils.Utilities
import utils.ComputeResults
import java.io.PrintWriter
import org.apache.spark.rdd.MapPartitionsRDD
import org.apache.spark.rdd.RDD
import scala.util.Random

import org.apache.spark.mllib.tree.DecisionTree
import org.apache.spark.mllib.tree.model.DecisionTreeModel
import org.apache.spark.mllib.tree.{impurity, DecisionTree, RandomForest}
import org.apache.spark.mllib.tree.configuration.{Algo, Strategy}
import org.apache.spark.mllib.tree.configuration.Algo._
import org.apache.spark.mllib.util.MLUtils

import org.apache.spark.mllib.classification.SVMModel
import org.apache.spark.mllib.classification.SVMWithSGD
import org.apache.spark.mllib.tree.RandomForest
import org.apache.spark.mllib.tree.model.RandomForestModel
import org.apache.mahout.keel.Algorithms.ImbalancedClassification.Ensembles.EUS

/**
 * @author Isaac Triguero
 * 
 * This class implements RUS and EUS undersampling approaches + classification (RF, DT, SVM) 
 */
object UndersamplingExample {
 
  def main(arg: Array[String]) {
         
    Logger.getLogger("org").setLevel(Level.OFF)
    
    var logger = Logger.getLogger(this.getClass())
    
    if (arg.length < 7) {
      logger.error("=> wrong parameters number")
      System.err.println("Parameters \n\t<path-to-header>\n\t<path-to-train>\n\t<path-to-test>\n\t<number-of-partitionUndersampling>\n\t<number-of-partitionClassifer>\n\t<preprocessingModel>\n\t<classificationModel>\n\t<pathOutput>")
      System.exit(1)
    }

    //Reading parameters
    val pathHeader = arg(0)
    val pathTrain = arg(1)
    val pathTest = arg(2)
    val numPartitionPreprocessing = arg(3).toInt
    val numPartitionClassification = arg(4).toInt
    val preprocessingMethod = arg(5)
    var classificationMethod = arg(6)
    val pathOutput = arg(7)
    
    //Basic Spark setup
    val jobName = "Undersampling Spark" + "-" + preprocessingMethod+"("+numPartitionPreprocessing+")" + "-"+ classificationMethod +"("+numPartitionClassification+")" 
    val conf = new SparkConf().setAppName(jobName).setMaster("local[*]")
    val sc = new SparkContext(conf)
    
    val timeStartModel = System.nanoTime  // take initial time.

        
    // Obtain the header.
    var HeaderString = ""
     for(x <- sc.textFile(pathHeader).collect())  HeaderString = HeaderString+"\n"+ x
     
    // Read the training dataset with numPartitionPreprocessing. 
    val trainRaw = sc.textFile(pathTrain: String, numPartitionPreprocessing).persist() // we use persist to re-use this variable.

    print("trainRaw.count(): " + trainRaw.count()+"\n") 
    
    // Here some code to read the header file which will be used to transform the data into LabeledPoints (Spark API).
    val typeConversion = KeelParser.parseHeaderFile(sc, pathHeader)
    val bcTypeConv = sc.broadcast(typeConversion)
    
    
    // Decision Tree and RandomForest parameters:
    val algo: Algo = Classification
    
    val strategy = new Strategy(algo = algo,
          impurity = impurity.Entropy,
          maxDepth = 5,  // Maximum allowed is 30
          maxBins = 1000,
          numClasses = 2,
          minInstancesPerNode = 2,
          minInfoGain = 0
         )
    
    // initialising partial runtimes.
    var PreprocessingTime = 0.0
    var LearningTime = 0.0
    var ClassificationTime = 0.0
    
    // Preprocessing Stage:
    var timeStartPreprocessing =System.nanoTime
    
    val data = if (preprocessingMethod.equalsIgnoreCase("RUS")) {
        trainRaw.mapPartitions(dataset => RandomUndersamplingMixedDataset(dataset)).persist()       
    }else if (preprocessingMethod.equalsIgnoreCase("EUS")) {
        var param = "seed = 48127491\n" + "ISmethod = QstatEUB_M_GM\n" + "No. of window minority = 1\n" + "No. of window majority = -1"; 
        trainRaw.mapPartitions(dataset => preprocessingEUS(dataset, param, HeaderString)).persist() 
    }else{ 
        trainRaw // if you don't want to apply any preprocessing.
    }
        
    logger.info("=> Size of the preprocessed dataset: \"" + data.count() + "\"")
    
    // coalescing if necessary.
    val dataRepartitioned = if(numPartitionClassification!= numPartitionPreprocessing){
       data.coalesce(numPartitionClassification).persist()
    }else{
       data
    }
           
    val endPreprocessing = System.nanoTime
      
    PreprocessingTime = (endPreprocessing - timeStartPreprocessing)/ 1e9
    
    // Learning phase:
    
    var modelDT:DecisionTreeModel = null
    var modelRF:RandomForestModel = null
    var modelSVM:SVMModel = null
    
    var timeStartLearning = System.nanoTime
      
    // converting the resulting dataset to RDD[LabeledPoint]
    val train = dataRepartitioned.map(line => KeelParser.parseLabeledPoint(bcTypeConv.value, line)).cache
      
    print("train.count(): " + train.count()+"\n") 
      
    dataRepartitioned.unpersist()  // we can delete the  
      
    if (classificationMethod.equalsIgnoreCase("DecisionTree")) {
        modelDT = DecisionTree.train(train,strategy)
    }else if (classificationMethod.equalsIgnoreCase("RandomForest")) {
        modelRF = RandomForest.trainClassifier(train,strategy,100,"sqrt",12345678)
    }else if (classificationMethod.equalsIgnoreCase("SVM")) {
        modelSVM = SVMWithSGD.train(train,100)
    }
           
    val endLearning = System.nanoTime
    LearningTime = (endLearning - timeStartLearning)/ 1e9
    
    data.unpersist()
    
    //  Classification of test set:
     
    val timeStartClassification = System.nanoTime

    // read test data
    val testRaw = sc.textFile(pathTest: String, numPartitionClassification)   
    val testData = testRaw.map(line => KeelParser.parseLabeledPoint(bcTypeConv.value, line)).cache
    
    val labelsTestData = testData.map { point =>
       point.label.toInt
      }.collect()                                      // take the label for evaluation
      
    val sizeTest= labelsTestData.length
  
    print("sizeTest: "+ sizeTest + "\n")
            
    val Preds =  if (classificationMethod.equalsIgnoreCase("DecisionTree")) {
          testData.map { point =>
          val prediction = modelDT.predict(point.features)
          (prediction.toInt)
          }.collect() 
    }else if (classificationMethod.equalsIgnoreCase("RandomForest")) {
          testData.map { point =>
          val prediction = modelRF.predict(point.features)
          (prediction.toInt)
          }.collect()  
    }else  {
          testData.map { point =>
          val prediction = modelSVM.predict(point.features)
          (prediction.toInt)
           }.collect()  
     }          

    val timeEndClassification = System.nanoTime

    ClassificationTime = ClassificationTime + (timeEndClassification - timeStartClassification)/ 1e9
    
    
   // printing out the results
    
    var writerResult = new String

    var writerResultALL:Array[String] = new Array[String](3)
    
    writerResult += "##########################################################\n"
    writerResult += "Confusion Matrix\n"
    var matrix = calculateConfusionMatrix(labelsTestData, Preds,2)
    matrix.foreach { x => writerResult += (x(0) + "\t\t\t\t" + x(1)) + "\n" }
    writerResult += "##########################################################\n"
    writerResult += "AUC:\t\t" + ComputeResults.computeAuc(matrix.toArray).toString() + "\n"
    writerResult += "GM:\t\t" + ComputeResults.computeGM(matrix.toArray).toString() + "\n"
    writerResult += "##########################################################\n"
    
    writerResult += "Preprocessing Time:\t\t" + PreprocessingTime + " seconds" + "\n"
    writerResult += "Learning Time:\t\t" + LearningTime + " seconds" + "\n"
    writerResult += "Classification Time:\t\t" + ClassificationTime + " seconds" + "\n"
    
    logger.info(writerResult)
    val pw = new PrintWriter(new File(pathOutput + "/Result.out")) 
    pw.write("\n" + writerResult)
    pw.close     
    
    
  }

  
  /**
   * This function perform (a transformation) the preprocessing of the original training data. It provides a new RDD.
   */

  def preprocessingEUS[T](iter: Iterator[String], param: String, header: String): Iterator[String] = {

    val (iterator1, iterator2) = iter.duplicate  // I duplicate it because I will need to go through the data twice.
    
    var logger = Logger.getLogger(this.getClass())
    var parameters = new parseParameters()
    parameters.parseConfigurationString(param);

    Utils.readHeader(header);

    var auxSet = new ArrayList[String]

    var IS = new InstanceSet()

    //Parser 
    while (iterator1.hasNext) {
      val cur = iterator1.next
      auxSet.add(cur.toString())
    }
    //logger.info(auxSet.toString()+"\n")
    IS.readSet(auxSet, true)

    // Conection:
    var eus = new EUS(IS, parameters,null,0)
    
    var seleccionadas = eus.getSelection()
    val selected = iterator2.zipWithIndex.filter{case (key, value) => seleccionadas(value.toInt)}

    val resulting =selected.map(line => line._1)

    resulting
  }



  def RandomUndersamplingMixedDataset[T](iter: Iterator[String]): Iterator[String] = {

    val r = scala.util.Random
    
    val (iterator1, iterator2) = iter.duplicate  // I duplicate it because I will need to go through the data twice.

    var iter_positive = iterator1.filter(line => line.contains("positive"))
    var iter_negative = iterator2.filter(line => line.contains("negative")) 
    
    val (iter_negative1, iter_negative2) = iter_negative.duplicate
    val (iter_positive1, iter_positive2) = iter_positive.duplicate
    
    val positive = iter_positive1.size
    val negative = iter_negative1.size
    
    var aux1:Array[Boolean]=Array.fill(positive){true}
    var aux2:Array[Boolean]=Array.fill(negative-positive){false}
    
    var selection = Array.concat(aux1,aux2)
    var randomSelection = util.Random.shuffle(selection.toList)    
    
    val selected = iter_negative2.zipWithIndex.filter{case (key, value) => randomSelection(value.toInt)}
    
    val otra =selected.map(line => line._1)

    otra++iter_positive2  // concatenate the selected negatives with the original positive.
  }

  
  /**
   * @brief Calculate the confusion matrix
   *
   * @param rightPredictedClas Array of int with right and predicted class
   * @param numClass Number of label for the objective class
   *
   */
  def calculateConfusionMatrix(OriginalLabels:Array[Int], predictions:Array[Int], numClass: Int): Array[Array[Int]] = {
    //Create and initializate the confusion matrix
    var confusionMatrix = Array.fill(numClass,numClass)(0)
      
    var size = OriginalLabels.length
    
    for (j <- 0 until size) {
        confusionMatrix(OriginalLabels(j))(predictions(j)) = confusionMatrix(OriginalLabels(j))(predictions(j)) + 1
    }

    confusionMatrix
  }
  
  
}

  
